from quart import Quart, jsonify, Response, Request

app = Quart(__name__)


@app.get("/")
async def index() -> Response:
    return jsonify({"message": "Hello, world!"})


@app.post("/foo")
async def post_foo() -> Response:
    resp = jsonify({"foo": "bar"})
    resp.headers["X-Test"] = "Test"
    return resp


@app.get("/exception")
async def mock_exception() -> Response:
    raise Exception("Mock exception")


if __name__ == "__main__":
    app.run()



from typing import Any

import pytest


@pytest.fixture
def base_config() -> dict[str, Any]:
    return {
        "url": "https://google.com",
        "body": {"data": "content"},
        "queryParams": {"foo": "bar"},
        "headers": {"Authorization": "TOKEN"},
    }



from typing import List

import pytest

from src.loamy.session import Clump, RequestMap, RequestResponse


@pytest.fixture(scope="session")
def request_map_collection() -> List[RequestMap]:
    requests: List[RequestMap] = []
    for i in range(0, 100):
        if i % 2 == 0:
            requests.append(
                RequestMap(
                    url="http://localhost:44777/",
                    http_op="GET",
                )
            )
        else:
            requests.append(
                RequestMap(
                    url="http://localhost:44777/foo",
                    http_op="POST",
                    body={"foo": "bar"},
                )
            )
    return requests


@pytest.fixture(scope="session")
def request_map_to_trigger_exception() -> RequestMap:
    return RequestMap(
        url="http://localhost:44777/exception",
        http_op="GET",
    )


def test_send_requests(request_map_collection: List[RequestMap]) -> None:
    session = Clump(requests=request_map_collection)
    responses: List[RequestResponse] = session.send_requests()
    assert len(responses) == 100
    for response in responses:
        assert response.status_code == 200
        assert response.error is None
        assert response.headers is not None
        assert response.headers["Content-Type"] == "application/json"
        if response.request_map.http_op == "POST":
            assert response.headers is not None
            assert "x-test" in response.headers
            assert "Test" == response.headers["x-test"]


def test_send_requests_with_exceptions(
    request_map_collection: List[RequestMap],
    request_map_to_trigger_exception: RequestMap,
) -> None:
    requests: List[RequestMap] = request_map_collection.copy()
    requests.append(request_map_to_trigger_exception)
    session = Clump(requests=requests)
    responses: List[RequestResponse] = session.send_requests(return_exceptions=True)
    assert len(responses) == 101
    for response in responses:
        if response.request_map.url == "http://localhost:44777/exception":
            assert response.status_code == 500
            assert response.error is not None
        else:
            assert response.status_code == 200
            assert response.error is None



import asyncio
from json import JSONDecodeError
from typing import Literal

import aiohttp
from loguru import logger
from pydantic import BaseModel, ConfigDict

# Disable the logger. If a consuming app wishes to see loamy's logs, they can enable() it again.
logger.disable("loamy")
# https://loguru.readthedocs.io/en/stable/overview.html#suitable-for-scripts-and-libraries
try:
    import uvloop

    async_run = uvloop.run
    logger.debug("Using uvloop for async operations")
except ModuleNotFoundError:
    async_run = asyncio.run  # type: ignore
    logger.debug("Using asyncio for async operations")


class RequestMap(BaseModel):
    """
    Class containing information about a single HTTP request to be sent.
    """

    url: str
    http_op: Literal["GET", "POST", "PUT", "PATCH", "OPTIONS", "DELETE"]
    body: dict | None = None
    query_params: dict[str, str] | None = None
    headers: dict[str, str] | None = None


class RequestResponse(BaseModel):
    """
    Class containing information about the result of an HTTP request.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    request_map: RequestMap
    status_code: int
    body: dict | None = None
    headers: dict[str, str] | None = None
    error: BaseException | None = None


class Clump:
    """
    Class for sending multiple HTTP requests concurrently.
    """

    def __init__(self, requests: list[RequestMap]) -> None:
        self._requestMaps: list[RequestMap] = requests
        logger.debug(f"Clump created with {len(self._requestMaps)} requests")

    def send_requests(self, return_exceptions: bool = False) -> list[RequestResponse]:
        logger.info(
            f"Sending {len(self._requestMaps)!s} requests with {return_exceptions=!s}"
        )
        return async_run(self._send_requests(rtn_exc=return_exceptions))

    async def _send_requests(self, rtn_exc: bool) -> list[RequestResponse]:
        async with aiohttp.ClientSession() as session:
            http_tasks: list[asyncio.Task] = [
                asyncio.ensure_future(self._route_request(req, session))
                for req in self._requestMaps
            ]
            logger.debug("Beginnging execution of request coroutines")
            results: list[RequestResponse | BaseException] = await asyncio.gather(
                *http_tasks, return_exceptions=rtn_exc
            )
            logger.debug("Finished execution of request coroutines")
            responses: list[RequestResponse] = []
            for i, resp in enumerate(results):
                if isinstance(resp, BaseException):
                    logger.error(f"Manually creating RequestResponse object and nesting error - index {i}")  # fmt: skip
                    responses.append(
                        RequestResponse(
                            request_map=self._requestMaps[i],
                            status_code=418,
                            error=resp,
                        )
                    )
                else:
                    responses.append(resp)
            logger.info(f"Returning {len(responses)!s} responses")
            return responses

    async def _route_request(
        self, req_map: RequestMap, session: aiohttp.ClientSession
    ) -> RequestResponse:
        response: RequestResponse = RequestResponse(request_map=req_map, status_code=0)
        try:
            logger.debug(f"Sending {req_map.http_op} request to {req_map.url}")
            match req_map.http_op:
                case "GET":
                    response = await self._send_get_request(req_map, session)
                case "POST":
                    response = await self._send_post_request(req_map, session)
                case "PUT":
                    response = await self._send_put_request(req_map, session)
                case "PATCH":
                    response = await self._send_patch_request(req_map, session)
                case "OPTIONS":
                    response = await self._send_options_request(req_map, session)
                case "DELETE":
                    response = await self._send_delete_request(req_map, session)
                case _:
                    logger.error("No matching HTTP operation found")
                    raise NotImplementedError
        except Exception as e:
            logger.exception(
                f"Error sending {req_map.http_op} request to {req_map.url}"
            )
            response.error = e

        return response

    async def _send_get_request(
        self, req_map: RequestMap, session: aiohttp.ClientSession
    ) -> RequestResponse:
        async with session.get(
            req_map.url, headers=req_map.headers, params=req_map.query_params
        ) as resp:
            error: aiohttp.ContentTypeError | JSONDecodeError | None = None
            status_code: int = resp.status
            logger.debug(f"{resp.url} returned {status_code}")
            try:
                body = await resp.json()
            except (aiohttp.ContentTypeError, JSONDecodeError) as e:
                logger.error(f"Failed to decode JSON response from {resp.url}")
                error = e
                logger.trace("Attempting to read response as text")
                text: str = await resp.text()
                logger.trace("Successfully read response as text")
                body = {"text": text}
        response = RequestResponse(
            request_map=req_map,
            status_code=status_code,
            body=body,
            headers=dict(resp.headers),
            error=error,
        )
        return response

    async def _send_post_request(
        self, req_map: RequestMap, session: aiohttp.ClientSession
    ) -> RequestResponse:
        async with session.post(
            req_map.url,
            json=req_map.body,
            headers=req_map.headers,
            params=req_map.query_params,
        ) as resp:
            error: aiohttp.ContentTypeError | JSONDecodeError | None = None
            status_code: int = resp.status
            logger.debug(f"{resp.url} returned {status_code}")
            try:
                body = await resp.json()
            except (aiohttp.ContentTypeError, JSONDecodeError) as e:
                logger.error(f"Failed to decode JSON response from {resp.url}")
                error = e
                logger.trace("Attempting to read response as text")
                text: str = await resp.text()
                logger.trace("Successfully read response as text")
                body = {"text": text}
        response = RequestResponse(
            request_map=req_map,
            status_code=status_code,
            body=body,
            headers=dict(resp.headers),
            error=error,
        )
        return response

    async def _send_put_request(
        self, req_map: RequestMap, session: aiohttp.ClientSession
    ) -> RequestResponse:
        async with session.put(
            req_map.url,
            json=req_map.body,
            headers=req_map.headers,
            params=req_map.query_params,
        ) as resp:
            status_code: int = resp.status
            error: aiohttp.ContentTypeError | JSONDecodeError | None = None
            logger.debug(f"{resp.url} returned {status_code}")
            try:
                body = await resp.json()
            except (aiohttp.ContentTypeError, JSONDecodeError) as e:
                logger.error(f"Failed to decode JSON response from {resp.url}")
                error = e
                logger.trace("Attempting to read response as text")
                text: str = await resp.text()
                logger.trace("Successfully read response as text")
                body = {"text": text}
        response = RequestResponse(
            request_map=req_map,
            status_code=status_code,
            body=body,
            headers=dict(resp.headers),
            error=error,
        )
        return response

    async def _send_patch_request(
        self, req_map: RequestMap, session: aiohttp.ClientSession
    ) -> RequestResponse:
        async with session.patch(
            req_map.url,
            json=req_map.body,
            headers=req_map.headers,
            params=req_map.query_params,
        ) as resp:
            status_code: int = resp.status
            error: aiohttp.ContentTypeError | JSONDecodeError | None = None
            logger.debug(f"{resp.url} returned {status_code}")
            try:
                body = await resp.json()
            except (aiohttp.ContentTypeError, JSONDecodeError) as e:
                logger.error(f"Failed to decode JSON response from {resp.url}")
                error = e
                logger.trace("Attempting to read response as text")
                text: str = await resp.text()
                logger.trace("Successfully read response as text")
                body = {"text": text}
        response = RequestResponse(
            request_map=req_map,
            status_code=status_code,
            body=body,
            headers=dict(resp.headers),
            error=error,
        )
        return response

    async def _send_options_request(
        self, req_map: RequestMap, session: aiohttp.ClientSession
    ) -> RequestResponse:
        async with session.options(
            req_map.url,
            json=req_map.body,
            headers=req_map.headers,
            params=req_map.query_params,
        ) as resp:
            status_code: int = resp.status
            error: aiohttp.ContentTypeError | JSONDecodeError | None = None
            logger.debug(f"{resp.url} returned {status_code}")
            try:
                body = await resp.json()
            except (aiohttp.ContentTypeError, JSONDecodeError) as e:
                logger.error(f"Failed to decode JSON response from {resp.url}")
                error = e
                logger.trace("Attempting to read response as text")
                text: str = await resp.text()
                logger.trace("Successfully read response as text")
                body = {"text": text}
        response = RequestResponse(
            request_map=req_map,
            status_code=status_code,
            body=body,
            headers=dict(resp.headers),
            error=error,
        )
        return response

    async def _send_delete_request(
        self, req_map: RequestMap, session: aiohttp.ClientSession
    ) -> RequestResponse:
        async with session.delete(
            req_map.url,
            json=req_map.body,
            headers=req_map.headers,
            params=req_map.query_params,
        ) as resp:
            status_code: int = resp.status
            error: aiohttp.ContentTypeError | JSONDecodeError | None = None
            logger.debug(f"{resp.url} returned {status_code}")
            try:
                body = await resp.json()
            except (aiohttp.ContentTypeError, JSONDecodeError) as e:
                logger.error(f"Failed to decode JSON response from {resp.url}")
                error = e
                logger.trace("Attempting to read response as text")
                text: str = await resp.text()
                logger.trace("Successfully read response as text")
                body = {"text": text}
        response = RequestResponse(
            request_map=req_map,
            status_code=status_code,
            body=body,
            headers=dict(resp.headers),
            error=error,
        )
        return response



#  ruff: noqa: S101, N802
from typing import Any

from src.loamy.session import RequestMap


def test_valid_RequestMap(base_config: dict[str, Any]):
    req_map = RequestMap(
        url=base_config["url"],
        http_op="GET",
        body=base_config["body"],
        query_params=base_config["queryParams"],
        headers=base_config["headers"],
    )
    assert req_map.url == base_config["url"]
    assert req_map.http_op == "GET"
    assert req_map.body == base_config["body"]
    assert req_map.query_params == base_config["queryParams"]
    assert req_map.headers == base_config["headers"]



# from loamy.session import Clump, RequestMap
# from typing import List
# import time
# import httpx

# requestMap = RequestMap(
#         url="http://localhost:8080/albums",
#         http_op="GET",
#     )

# sampleRequests: List[RequestMap] = []

# for i in range(5000):
#     sampleRequests.append(requestMap)


# def test_async_requests():
#     session = Clump(requests=sampleRequests)
#     startTime = time.time()
#     responses = session.send_requests()
#     endTime = time.time()
#     elapsedTime = endTime - startTime
#     print(f"Asynchronous operation completed {len(sampleRequests)} requests in {elapsedTime} seconds")
#     assert len(responses) != 0
#     assert elapsedTime < 1

# def test_sync_requests():
#     client = httpx.Client(transport=httpx.HTTPTransport(local_address="0.0.0.0"))
#     startTime = time.time()
#     for request in sampleRequests:
#         client.get(request.url)
#     endTime = time.time()
#     elapsedTime = endTime - startTime
#     print(f"Synchronous operation completed {len(sampleRequests)} requests in {elapsedTime} seconds")
#     assert elapsedTime < 3
